/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.yoko.bindings.corba;

import java.io.IOException;
import java.lang.ref.WeakReference;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executor;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Port;
import javax.wsdl.WSDLException;
import javax.xml.namespace.QName;
import javax.xml.ws.WebServiceException;

import org.apache.schemas.yoko.bindings.corba.AddressType;
import org.apache.schemas.yoko.bindings.corba.Exception;
import org.apache.schemas.yoko.bindings.corba.MemberType;
import org.apache.schemas.yoko.bindings.corba.OperationType;
import org.apache.schemas.yoko.bindings.corba.RaisesType;
import org.apache.yoko.bindings.corba.types.CorbaExceptionHandler;
import org.apache.yoko.bindings.corba.types.CorbaHandlerUtils;
import org.apache.yoko.bindings.corba.types.CorbaObjectHandler;

import org.objectweb.celtix.Bus;
import org.objectweb.celtix.bindings.AbstractBindingBase;
import org.objectweb.celtix.bindings.AbstractBindingImpl;
import org.objectweb.celtix.bindings.ClientBinding;
import org.objectweb.celtix.bindings.DataBindingCallback;
import org.objectweb.celtix.bindings.ResponseCallback;
import org.objectweb.celtix.buslifecycle.BusLifeCycleListener;
import org.objectweb.celtix.common.logging.LogUtils;
import org.objectweb.celtix.context.ObjectMessageContext;
import org.objectweb.celtix.ws.addressing.EndpointReferenceType;
import org.objectweb.celtix.wsdl.EndpointReferenceUtils;

import org.omg.CORBA.Any;
import org.omg.CORBA.Context;
import org.omg.CORBA.ContextList;
import org.omg.CORBA.ExceptionList;
import org.omg.CORBA.NVList;
import org.omg.CORBA.NamedValue;
import org.omg.CORBA.ORB;
import org.omg.CORBA.Request;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.UnknownUserException;

public class CorbaClientBinding extends AbstractBindingBase implements ClientBinding {

    private static final Logger LOG = LogUtils.getL7dLogger(CorbaClientBinding.class);

    protected final CorbaBindingImpl corbaBinding;
    protected EndpointReferenceType endpointRef;
    protected ORB orb;
    protected org.omg.CORBA.Object target;

    private class CorbaBusLifeCycleListener extends WeakReference<CorbaClientBinding> 
        implements BusLifeCycleListener {

        CorbaBusLifeCycleListener(CorbaClientBinding c) {
            super(c);
        }
    
        public void initComplete() {
            // Nothing needed for the binding
        }
    
        public void preShutdown() {
            // Nothing needed for the binding
        }
    
        public void postShutdown() {
            // TODO: Revisit: Is this being invoked?
            if (orb != null) {
                try {
                    orb.destroy();
                } catch (java.lang.Exception ex) {
                    // TODO: For now, we just consume.  Revisit.
                }
            }
        }
    }

    public CorbaClientBinding(Bus b, EndpointReferenceType ref)
        throws WSDLException, IOException {
        super(b, ref);
        endpointRef = ref;
        
        bus.getLifeCycleManager().registerLifeCycleListener(new CorbaBusLifeCycleListener(this));
        
        // TODO: Set any additional properties needed to initialize the ORB  before
        // we initialize it. We will get this information from the Celtix configuration 
        // that is used with the current application.
        java.util.Properties props = System.getProperties();
        props.put("org.omg.CORBA.ORBClass", "org.apache.yoko.orb.CORBA.ORB");
        props.put("org.omg.CORBA.ORBSingletonClass",
                "org.apache.yoko.orb.CORBA.ORBSingleton");
        props.put("yoko.orb.id", "Yoko-Client-Binding");

        orb = ORB.init(new String[0], props);
        if (orb == null) {
            LOG.severe("Could not create instance of the ORB");
            throw new WSDLException(WSDLException.OTHER_ERROR,
                    "Could not create instance of the ORB");
        }

        try {
            // obtain a reference to the target object
            Port wsdlPort = EndpointReferenceUtils.getPort(b.getWSDLManager(),
                    this.endpointRef);

            List extElements = wsdlPort.getExtensibilityElements();
            AddressType address = null;
            for (Iterator iter = extElements.iterator(); iter.hasNext();) {
                Object e = iter.next();
                if (e instanceof AddressType) {
                    address = (AddressType) e;
                    break;
                }
            }

            if (address == null) {
                LOG.log(Level.SEVERE, "Unable to locate a valid CORBA address");
                throw new CorbaBindingException("Unable to locate a valid CORBA address");
            }
            
            if (!isBindingCompatible(address.getLocation())) {
                LOG.log(Level.SEVERE, "Address not a valid CORBA address");
                throw new CorbaBindingException("Address not a valid CORBA address");
            }

            target = CorbaUtils.importObjectReference(orb, address.getLocation());
        } catch (java.lang.Exception ex) {
            LOG.log(Level.SEVERE, "Could not resolve target object");
            throw new WSDLException(WSDLException.OTHER_ERROR, "Could not resolve target object", ex);
        }

        corbaBinding = new CorbaBindingImpl(b, ref, orb, false);
    }

    // -- from BindingBase interface --
    public AbstractBindingImpl getBindingImpl() {
        return corbaBinding;
    }

    // -- from ClientBinding interface --
    public boolean isBindingCompatible(String address) {
        // TODO - add support for file that contain IORs
        return address.startsWith("IOR:") 
               || address.startsWith("corbaloc:") 
               || address.startsWith("relfile:") 
               || address.startsWith("file:");
    }

    public ResponseCallback createResponseCallback() {
        // TODO - provide implementation
        return null;
    }

    public ObjectMessageContext invoke(ObjectMessageContext objectCtx,
            DataBindingCallback callback) throws IOException {
        CorbaMessageContext mc = (CorbaMessageContext) corbaBinding.createBindingMessageContext(objectCtx);
        mc.put(ObjectMessageContext.MESSAGE_INPUT, Boolean.TRUE);

        corbaBinding.marshal(objectCtx, mc, callback);
        CorbaMessage message = mc.getMessage();

        try {
            objectCtx.put(ObjectMessageContext.WSDL_OPERATION, callback.getOperationName());

            // Build the list of DII arguments, contexts, returns, and exceptions
            CorbaStreamable[] arguments = message.getStreamableArguments();
            NVList list = orb.create_list(arguments.length);

            for (int i = 0; i < arguments.length; ++i) {
                Any value = orb.create_any();
                value.insert_Streamable(arguments[i]);
                list.add_value(arguments[i].getName(), value, arguments[i].getMode());
            }

            ContextList ctxList = orb.create_context_list();
            Context ctx = orb.get_default_context();

            CorbaStreamable retVal = mc.getMessage().getStreamableReturn();
            NamedValue ret = null;
            if (retVal != null) {
                Any returnAny = orb.create_any();
                returnAny.insert_Streamable(retVal);
                ret = orb.create_named_value(retVal.getName(), returnAny, org.omg.CORBA.ARG_OUT.value);
            } else {
                // TODO: REVISIT: for some reason, the yoko ORB does not like to have a null NamedValue
                // return value. Create this 'empty' one if a void return type is used.
                ret = orb.create_named_value("return", orb.create_any(), org.omg.CORBA.ARG_OUT.value);
            }

            // Get the typecodes for the exceptions this operation can throw.  These are defined in the
            // operation definition from WSDL.
            ExceptionList exList = orb.create_exception_list();
            OperationType operation = CorbaUtils.getCorbaOperationType(callback.getOperationName(), 
                                                                       bus, 
                                                                       endpointRef);

            if (operation == null) {
                throw new CorbaBindingException("Unable to obtain operation definition");
            }

            Map<TypeCode, RaisesType> exceptions = getOperationExceptions(operation);
            Object[] tcs = exceptions.keySet().toArray();
            for (int i = 0; i < exceptions.size(); ++i) {
                exList.add((TypeCode) tcs[i]);
            }

            Request request = 
                target._create_request(ctx, 
                                       (String)objectCtx.get(ObjectMessageContext.WSDL_OPERATION), 
                                       list, 
                                       ret,
                                       exList, 
                                       ctxList);
            request.invoke();

            mc.put(ObjectMessageContext.MESSAGE_INPUT, Boolean.FALSE);
            java.lang.Exception ex = request.env().exception();
            if (ex == null) {
                if (retVal != null) {
                    message.setStreamableReturnValue(retVal.getObject());
                }

                for (int i = 0; i < list.count(); ++i) {
                    NamedValue arg = (NamedValue) list.item(i);
                    if (arg.flags() != org.omg.CORBA.ARG_IN.value) {
                        CorbaStreamable streamable = (CorbaStreamable)arg.value().extract_Streamable();
                        message.setStreamableArgumentValue(streamable.getObject(), i);
                    }
                }
                corbaBinding.unmarshal(mc, objectCtx, callback);
            } else if (ex instanceof UnknownUserException) {
                UnknownUserException userEx = (UnknownUserException)ex;
                Any exAny = userEx.except;
                RaisesType raisesExType = exceptions.get(exAny.type());
                
                CorbaStreamable streamableEx = createStreamableException(exAny, raisesExType);
                mc.getMessage().setStreamableException(streamableEx);
                corbaBinding.unmarshalFault(mc, objectCtx, callback);
            } else {
                objectCtx.setException(new WebServiceException("Unknown CORBA return value"));
            }
        } catch (java.lang.Exception ex) {
            throw new CorbaBindingException("Exception during invoke", ex);
        }

        return objectCtx;
    }
    
    public void invokeOneWay(ObjectMessageContext objectCtx, DataBindingCallback callback) 
        throws IOException {
        CorbaMessageContext mc = (CorbaMessageContext)corbaBinding.createBindingMessageContext(objectCtx);
        mc.put(ObjectMessageContext.MESSAGE_INPUT, Boolean.TRUE);

        corbaBinding.marshal(objectCtx, mc, callback);
        CorbaMessage message = mc.getMessage();

        try {
            CorbaStreamable[] arguments = message.getStreamableArguments();
            NVList list = orb.create_list(arguments.length);

            for (int i = 0; i < arguments.length; ++i) {
                Any value = orb.create_any();
                value.insert_Streamable(arguments[i]);
                list.add_value(arguments[i].getName(), value, arguments[i].getMode());
            }

            ContextList ctxList = orb.create_context_list();
            Context ctx = orb.get_default_context();

            // TODO: This doen't really apply to a oneway since we are not expecting a return.
            CorbaStreamable retVal = mc.getMessage().getStreamableReturn();
            NamedValue ret = null;
            if (retVal != null) {
                Any returnAny = orb.create_any();
                returnAny.insert_Streamable(retVal);
                ret = orb.create_named_value(retVal.getName(), returnAny, org.omg.CORBA.ARG_OUT.value);
            } else {
                // TODO: REVISIT: for some reason, the yoko ORB does not like to have a null NamedValue
                // return value. Create this 'empty' one if a void return type is used.
                ret = orb.create_named_value("return", orb.create_any(), org.omg.CORBA.ARG_OUT.value);
            }

            // Get the typecodes for the exceptions this operation can throw.  These are defined in the
            // operation definition from WSDL.
            ExceptionList exList = orb.create_exception_list();
            OperationType operation = CorbaUtils.getCorbaOperationType(callback.getOperationName(), 
                                                                       bus, 
                                                                       endpointRef);

            if (operation == null) {
                throw new CorbaBindingException("Unable to obtain operation definition");
            }

            Map<TypeCode, RaisesType> exceptions = getOperationExceptions(operation);
            Object[] tcs = exceptions.keySet().toArray();
            for (int i = 0; i < exceptions.size(); ++i) {
                exList.add((TypeCode) tcs[i]);
            }

            Request request = 
                target._create_request(ctx, 
                                       (String)objectCtx.get(ObjectMessageContext.WSDL_OPERATION), 
                                       list, 
                                       ret,
                                       exList, 
                                       ctxList);
            request.send_oneway();
           
        } catch (java.lang.Exception ex) {
            throw new CorbaBindingException("Exception during invokeOneWay", ex);
        }
    }

    public Future<ObjectMessageContext> invokeAsync(ObjectMessageContext objectCtx, 
                                                    DataBindingCallback callback,
                                                    Executor executor) throws IOException {
        throw new CorbaBindingException("invokeAsync not yet implemented");
    }

    public Map<TypeCode, RaisesType> getOperationExceptions(
            OperationType operation) {

        Map<TypeCode, RaisesType> exceptions = new HashMap<TypeCode, RaisesType>();
        List<RaisesType> exList = operation.getRaises();

        for (Iterator<RaisesType> iter = exList.iterator(); iter.hasNext();) {
            RaisesType ex = iter.next();
            TypeCode tc = CorbaUtils.getTypeCode(orb, ex.getException(), corbaBinding.getCorbaTypeMaps());
            exceptions.put(tc, ex);
        }

        return exceptions;
    }

    public CorbaStreamable createStreamableException(Any exception, RaisesType exRaises) {
        // first, we need to get the definition of the exception.  All we have from the raises type is the 
        // exception idl type
        QName exIdlType = exRaises.getException();
        List<CorbaTypeMap> typeMaps = corbaBinding.getCorbaTypeMaps();
        Exception ex = null;
        for (Iterator i = typeMaps.iterator(); i.hasNext();) {
            CorbaTypeMap tm = (CorbaTypeMap)i.next();
            if (tm.getTargetNamespace().equals(exIdlType.getNamespaceURI())) {
                ex = (Exception)tm.getType(exIdlType.getLocalPart()); 
                break;
            }
        }
        
        if (ex == null) {
            throw new CorbaBindingException("Unable to find definition for thrown exception");
        }
        
        TypeCode tc = CorbaUtils.getTypeCode(orb, exIdlType, typeMaps);
        // TODO: Revisit - What should the name of the object holder be?
        CorbaExceptionHandler obj = new CorbaExceptionHandler(new QName(ex.getName()), exIdlType, tc, ex);
        List<MemberType> exMembers = ex.getMember();
        for (Iterator i = exMembers.iterator(); i.hasNext();) {
            MemberType member = (MemberType)i.next();
            QName memberName = new QName(member.getName());
            QName memberType = member.getIdltype();
            CorbaObjectHandler memberObj = 
                CorbaHandlerUtils.createTypeHandler(orb, memberName, memberType, typeMaps);
            obj.addMember(memberObj);
        }

        CorbaStreamable exStreamable = new CorbaStreamable(obj, new QName("exception"));
        exStreamable._read(exception.create_input_stream());
        
        return exStreamable;
    }
}
