/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.yoko.bindings.corba;

import java.io.IOException;
import java.util.logging.Logger;

import javax.wsdl.WSDLException;

import org.apache.schemas.yoko.bindings.corba.AddressType;

import org.objectweb.celtix.Bus;
import org.objectweb.celtix.bindings.AbstractBindingBase;
import org.objectweb.celtix.bindings.AbstractBindingImpl;
import org.objectweb.celtix.bindings.DataBindingCallback;
import org.objectweb.celtix.bindings.ServerBinding;
import org.objectweb.celtix.bindings.ServerBindingEndpointCallback;
import org.objectweb.celtix.common.logging.LogUtils;
import org.objectweb.celtix.context.OutputStreamMessageContext;
import org.objectweb.celtix.ws.addressing.EndpointReferenceType;

import org.omg.CORBA.ORB;
import org.omg.CORBA.Policy;
import org.omg.PortableServer.POA;
import org.omg.PortableServer.POAHelper;
import org.omg.PortableServer.POAManager;

public class CorbaServerBinding extends AbstractBindingBase implements ServerBinding {

    private static final Logger LOG = LogUtils.getL7dLogger(CorbaBindingImpl.class);

    protected final CorbaBindingImpl corbaBinding;
    protected ServerBindingEndpointCallback sbeCallback;
    private ORB orb;

    public CorbaServerBinding(Bus b, EndpointReferenceType ref, ServerBindingEndpointCallback cbFactory) {
        super(b, ref);

        sbeCallback = cbFactory;

        String location = "";
        AddressType address = CorbaUtils.getCorbaAddressType(bus, ref);
        if (address != null) {
            location = address.getLocation();
        }
        LOG.info("Service address retrieved: " + location);

        // TODO: Get any configuration options for the ORB and set the appropriate properties.
        java.util.Properties props = System.getProperties();
        props.put("org.omg.CORBA.ORBClass", "org.apache.yoko.orb.CORBA.ORB");
        props.put("org.omg.CORBA.ORBSingletonClass", "org.apache.yoko.orb.CORBA.ORBSingleton");
        props.put("yoko.orb.id", "Yoko-Server-Binding");

        // If we have one of these two types of addresses, we should specify the address
        // and port information as properties to the ORB during initialization.
        if (location.startsWith("IOR:")) {
            // TODO: How to handle this? Can we obtain port/host information from the IOR
            // or do we just want to let the ORB pick the port and export the IOR to the 
            // Logs?
        } else if (location.startsWith("corbaloc")) {
            int corbalocIndex = location.indexOf(":");
            int protocolIndex = location.indexOf(":", corbalocIndex + 1);
            String protocol = location.substring(corbalocIndex + 1, protocolIndex);
            int hostIndex = location.indexOf(":", protocolIndex + 1);
            String host = location.substring(protocolIndex + 1, hostIndex);
            // the port number should be followed by a '/' character
            int portIndex = location.indexOf("/", hostIndex + 1);
            String port = location.substring(hostIndex + 1, portIndex);
            if (protocol.length() == 0) {
                // This means no protocol was defined.  Default to iiop
                protocol = "iiop";
            }
            props.put("yoko.orb.oa.endpoint", new String(protocol + " --host " + host + " --port " + port));
            LOG.info("Set server endpoint: " + protocol + " --host " + host + " --port " + port);
        }
        orb = ORB.init(new String[0], props);

        corbaBinding = new CorbaBindingImpl(b, ref, orb, true);
    }

    // -- from BindingBase interface --
    public AbstractBindingImpl getBindingImpl() {
        return corbaBinding;
    }

    public boolean isBindingCompatible(String address) {
        // TODO - add support for files which contain IORs
        return address.startsWith("IOR:") 
               || address.startsWith("corbaloc:") 
               || address.startsWith("relfile") 
               || address.startsWith("file");
    }

    // -- from ServerBinding interface --
    public void activate() throws WSDLException, IOException {
        // Get the address that the DSI servant should be publishing

        String location = "";
        AddressType address = CorbaUtils.getCorbaAddressType(corbaBinding.getBus(),
                                                             corbaBinding.getEndpointReference());
        if (address != null) {
            location = address.getLocation();
        }

        org.omg.CORBA.Object obj = null;
        try {
            POA rootPOA = POAHelper.narrow(orb.resolve_initial_references("RootPOA"));
            POAManager poaManager = rootPOA.the_POAManager();

            Policy[] policies = new Policy[3];
            policies[0] = rootPOA
                    .create_lifespan_policy(
                        org.omg.PortableServer.LifespanPolicyValue.PERSISTENT);
            policies[1] = rootPOA
                    .create_implicit_activation_policy(
                        org.omg.PortableServer.ImplicitActivationPolicyValue.NO_IMPLICIT_ACTIVATION);
            policies[2] = rootPOA
                    .create_id_uniqueness_policy(
                        org.omg.PortableServer.IdUniquenessPolicyValue.UNIQUE_ID);

            POA bindingPOA = rootPOA.create_POA("BindingPOA", poaManager, policies);
            CorbaDSIServant servant = new CorbaDSIServant(orb, bindingPOA, this, sbeCallback);
            byte[] objectId = bindingPOA.activate_object(servant);
            obj = bindingPOA.id_to_reference(objectId);
            
            if (location.startsWith("relfile:")) {
                String iorFile = location.substring("relfile:".length(), location.length());
                // allow for up to 3 '/' to match common uses of relfile url format
                for (int n = 0; n < 3; n++) {
                    if (iorFile.charAt(0) != '/') {
                        break;
                    } else {
                        iorFile = iorFile.substring(1);
                    }
                }

                CorbaUtils.exportObjectReferenceToFile(obj, orb, iorFile);
            } else if (location.startsWith("file:")) {
                String iorFile = location.substring("file:".length(), location.length());
                // allow for up to 3 '/' to match common uses of file url format
                for (int n = 0; n < 3; n++) {
                    if (iorFile.charAt(0) != '/') {
                        break;
                    } else {
                        iorFile = iorFile.substring(1);
                    }
                }
                // to match the ORB, file must have complete path information, therefore we add
                // a '/' prefix to the address
                //iorFile = "/" + iorFile;
                CorbaUtils.exportObjectReferenceToFile(obj, orb, iorFile);
            } else if (location.startsWith("corbaloc")) {
                // Try add the key to the boot manager.  This is required for a corbaloc
                try {
                    int keyIndex = location.indexOf('/');
                    String key = location.substring(keyIndex + 1);
                    org.apache.yoko.orb.OB.BootManager bootManager = 
                        org.apache.yoko.orb.OB.BootManagerHelper.narrow(
                            orb.resolve_initial_references("BootManager"));
                    bootManager.add_binding(key.getBytes(), obj);
                    LOG.info("Added key " + key + " to bootmanager");
                } catch (Exception ex) {
                    // TODO: Continue without for now
                }
                String ior = orb.object_to_string(obj);
                address.setLocation(ior);
                String iorFile = "endpoint.ior";
                CorbaUtils.exportObjectReferenceToFile(obj, orb, iorFile);
            } else {
                String ior = orb.object_to_string(obj);
                address.setLocation(ior);
                String iorFile = "endpoint.ior";
                CorbaUtils.exportObjectReferenceToFile(obj, orb, iorFile);
                LOG.info("Object Reference: " + orb.object_to_string(obj));
            }
            // TODO: Provide other export mechanisms? 
            
            poaManager.activate();
        } catch (Exception ex) {
            // TODO: Throw appropriate exception
            throw new CorbaBindingException("Unable to active CORBA servant", ex);
        }
    }

    public void deactivate() throws IOException {
        if (orb != null) {
            try {
                orb.destroy();
            } catch (Exception ex) {
                // TODO: For now, we just consume.  Revisit.
            }
        }
    }

    public void partialResponse(OutputStreamMessageContext outputContext,
                                DataBindingCallback callback) 
        throws IOException {
        // Complete? No concept of a partial response in CORBA
    }
}
