/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
*/

package org.apache.yoko.tools.processors.idl;

import java.util.Iterator;
import java.util.Map;

import javax.wsdl.Binding;
import javax.wsdl.Definition;
import javax.wsdl.PortType;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.ExtensionRegistry;

import javax.xml.namespace.QName;

import antlr.ASTVisitor;
import antlr.collections.AST;

import org.apache.schemas.yoko.bindings.corba.BindingType;

import org.apache.yoko.wsdl.CorbaConstants;

public class PortTypeVisitor implements ASTVisitor {   

    Definition definition;
    ExtensionRegistry extReg;
    PortType portType;
    WSDLASTVisitor wsdlASTVisitor;
    TypesVisitor typesVisitor;
    String module;

    public PortTypeVisitor(WSDLASTVisitor visitor) {
        definition = visitor.getDefinition();
        wsdlASTVisitor = visitor;
        typesVisitor = wsdlASTVisitor.getTypesVisitor();
        extReg = definition.getExtensionRegistry();
    }

    public void addModuleName(String name) {
        if (module != null) {
            module += "." + name;
        } else {
            module = name;
        }
    }

    public void visit(AST node) {
        portType = definition.createPortType();
        AST interfaceNode = node.getFirstChild();
        String intfName = interfaceNode.toString();
        if (module != null) {
            intfName = module + "." + intfName;
        }
        portType.setQName(new QName(definition.getTargetNamespace(), intfName));
        portType.setUndefined(false);
        definition.addPortType(portType);
        Binding binding = createBinding();
        AST node2 = interfaceNode.getNextSibling();
        while (node2 != null) {
            switch (node2.getType()) {
            case IDLTokenTypes.LITERAL_struct:
            case IDLTokenTypes.LITERAL_exception:
            case IDLTokenTypes.LITERAL_const:
            case IDLTokenTypes.LITERAL_enum:
            case IDLTokenTypes.LITERAL_union:
            case IDLTokenTypes.LITERAL_typedef:
                typesVisitor.visit(node2);
                break;
            default:
                OperationVisitor operationVisitor = new OperationVisitor(wsdlASTVisitor,
                                                                         portType,
                                                                         binding);
                operationVisitor.visit(node2);     
                break;
            }
            node2 = node2.getNextSibling();
        }
    }

    public Binding createBinding() {
        String bname = portType.getQName().getLocalPart() + "CORBABinding";
        QName bqname = new QName(definition.getTargetNamespace(),
                                 bname);
        int count = 0;
        while (queryBinding(bqname)) {
            bname = bname + count;
            bqname = new QName(definition.getTargetNamespace(), bname);
        }
        Binding binding = definition.createBinding();
        binding.setPortType(portType);
        binding.setQName(bqname);

        try {
            BindingType bindingType = (BindingType)
                extReg.createExtension(Binding.class, CorbaConstants.NE_CORBA_BINDING);
            bindingType.setRepositoryID(CorbaConstants.REPO_STRING
                                        + portType.getQName().getLocalPart().replace('.', '/')
                                        + CorbaConstants.IDL_VERSION);
            binding.addExtensibilityElement(bindingType);
        } catch (WSDLException ex) {
            throw new RuntimeException(ex);
        }
        binding.setUndefined(false);
        definition.addBinding(binding);
        return binding;
    }

    private boolean queryBinding(QName bqname) {
        Map bindings = definition.getBindings();
        Iterator i = bindings.values().iterator();
        while (i.hasNext()) {
            Binding binding = (Binding)i.next();
            if (binding.getQName().getLocalPart().equals(bqname.getLocalPart())) {
                return true;
            }
        }
        return false;
    }
}
