/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
*/

package org.apache.yoko.tools.processors.idl;

import javax.xml.namespace.QName;

import antlr.collections.AST;

import org.apache.schemas.yoko.bindings.corba.CaseType;
import org.apache.schemas.yoko.bindings.corba.TypeMappingType;
import org.apache.schemas.yoko.bindings.corba.Union;
import org.apache.schemas.yoko.bindings.corba.Unionbranch;

import org.apache.ws.commons.schema.XmlSchema;
import org.apache.ws.commons.schema.XmlSchemaChoice;
import org.apache.ws.commons.schema.XmlSchemaCollection;
import org.apache.ws.commons.schema.XmlSchemaComplexType;
import org.apache.ws.commons.schema.XmlSchemaElement;
import org.apache.ws.commons.schema.XmlSchemaSequence;
import org.apache.ws.commons.schema.XmlSchemaType;

import org.apache.yoko.wsdl.CorbaConstants;

public class UnionVisitor extends TypesVisitorBase {

    private final String discriminator = "discriminator";
    
    public UnionVisitor(XmlSchemaCollection xmlSchemas,
                       XmlSchema xmlSchema,
                       TypeMappingType typeMappingType) {
        super(xmlSchemas, xmlSchema, typeMappingType);
    }
    
    public void visit(AST unionNode) {
        AST identifierNode = unionNode.getFirstChild();
        AST discriminatorNode = identifierNode.getNextSibling();
        AST caseNode = discriminatorNode.getNextSibling();

        // xmlschema:union
        XmlSchemaComplexType unionSchemaComplexType = new XmlSchemaComplexType(schema);
        XmlSchemaSequence sequence = new XmlSchemaSequence();
        unionSchemaComplexType.setName(identifierNode.toString());
        unionSchemaComplexType.setParticle(sequence);
        
        XmlSchemaElement discriminatorElement = new XmlSchemaElement();
        XmlSchemaType type = TypesUtils.findType(schemas, schema, discriminatorNode);
        discriminatorElement.setName(discriminator);
        discriminatorElement.setSchemaTypeName(type.getQName());
        discriminatorElement.setMinOccurs(1);
        discriminatorElement.setMaxOccurs(1);
        sequence.getItems().add(discriminatorElement);
        
        XmlSchemaChoice choice = new XmlSchemaChoice();
        choice.setMinOccurs(0);
        choice.setMaxOccurs(1);
        sequence.getItems().add(choice);
        
        
        // corba:union
        Union corbaUnion = new Union();
        corbaUnion.setQName(new QName(typeMap.getTargetNamespace(), identifierNode.toString()));
        corbaUnion.setRepositoryID(CorbaConstants.REPO_STRING
                                   + identifierNode.toString()
                                   + CorbaConstants.IDL_VERSION);
        corbaUnion.setType(unionSchemaComplexType.getQName());
        corbaUnion.setDiscriminator(TypesUtils.findCorbaType(typeMap, type.getQName()).getQName());
        
        processCaseNodes(caseNode, choice, corbaUnion);
            
        
        // add xmlschema:union
        schema.getItems().add(unionSchemaComplexType);
        schema.addType(unionSchemaComplexType);

        // add corba:union
        typeMap.getStructOrExceptionOrUnion().add(corbaUnion);
    }
    
    private void processCaseNodes(AST caseNode,
                                  XmlSchemaChoice choice,
                                  Union corbaUnion) {
        while (caseNode != null) {
            AST typeNode  = null;
            AST nameNode  = null;
            AST labelNode = null;
            
            // xmlschema:element
            XmlSchemaElement element = new XmlSchemaElement();

            // corba:unionbranch
            Unionbranch unionBranch = new Unionbranch();

            if (caseNode.getType() == IDLTokenTypes.LITERAL_default) {
                // default:
                unionBranch.setDefault(true);
                
                typeNode = caseNode.getFirstChild();
                nameNode = typeNode.getNextSibling();
            } else {
                // case:
                createCase(caseNode, unionBranch);
                
                labelNode = caseNode.getFirstChild();
                if (labelNode.getType() == IDLTokenTypes.LITERAL_case) {
                    labelNode = labelNode.getNextSibling();
                }
                
                typeNode = labelNode.getNextSibling();
                nameNode = typeNode.getNextSibling();
            }
            
            // xmlschema:element
            element.setName(nameNode.toString());
            XmlSchemaType type = TypesUtils.findType(schemas, schema, typeNode);
            element.setSchemaTypeName(type.getQName());
            choice.getItems().add(element);
            
            
            // corba:unionbranch
            unionBranch.setName(nameNode.toString());
            QName idlType = TypesUtils.findCorbaType(typeMap, type.getQName()).getQName();
            unionBranch.setIdltype(idlType);
            corbaUnion.getUnionbranch().add(unionBranch);
            
            
            caseNode = caseNode.getNextSibling();
        }
    }
    
    private void createCase(AST caseNode, Unionbranch unionBranch) {
        AST node = caseNode.getFirstChild();
        if (node != null) {
            if (node.getType() == IDLTokenTypes.LITERAL_case) {
                // corba:case
                CaseType caseType = new CaseType();
                caseType.setLabel(node.getNextSibling().toString());
                unionBranch.getCase().add(caseType);
                
                // recursive call
                createCase(node, unionBranch);
            } else {
                // corba:case
                CaseType caseType = new CaseType();
                caseType.setLabel(node.toString());
                unionBranch.getCase().add(caseType);
            }
        }
    }
}
